@@ -6,7 +6,35 @@ typedef int Py_ssize_t;
 #define PY_SSIZE_T_MIN INT_MIN
 #endif
 
-#include <ImathBox.h>
+#if PY_MAJOR_VERSION >= 3
+  #define MOD_ERROR_VAL NULL
+  #define MOD_SUCCESS_VAL(val) val
+  #define MOD_INIT(name) PyMODINIT_FUNC PyInit_##name(void)
+  #define MOD_DEF(ob, name, doc, methods) \
+          static struct PyModuleDef moduledef = { \
+            PyModuleDef_HEAD_INIT, name, doc, -1, methods, }; \
+          ob = PyModule_Create(&moduledef);
+  #define PyInt_FromLong(x) PyLong_FromLong(x)
+  #define PyInt_AsLong(x) PyLong_AsLong(x)
+  #define PyInt_Check(x) PyLong_Check(x)
+  #define PyString_Check(x) PyBytes_Check(x)
+  #define PyString_AsString(x) PyBytes_AsString(x)
+  #define PyString_Size(x) PyBytes_Size(x)
+  #define PyString_FromString(x) PyBytes_FromString(x)
+  #define PyString_FromStringAndSize(x, y) PyBytes_FromStringAndSize(x, y)
+  #define PyUTF8_AsSstring(x)   PyString_AsString(PyUnicode_AsUTF8String(x))
+  #define PyUTF8_FromSstring(x)   Something...
+#else
+  #define MOD_ERROR_VAL
+  #define MOD_SUCCESS_VAL(val)
+  #define MOD_INIT(name) extern "C" void init##name(void)
+  #define MOD_DEF(ob, name, doc, methods) \
+          ob = Py_InitModule3(name, methods, doc);
+  #define PyUTF8_AsSstring(x)   PyString_AsString(x)
+#endif
+
+#include <ImfIO.h>
+#include <Iex.h>
 #include <ImfArray.h>
 #include <ImfAttribute.h>
 #include <ImfBoxAttribute.h>
@@ -18,6 +46,7 @@ typedef int Py_ssize_t;
 #include <ImfDoubleAttribute.h>
 #include <ImfEnvmapAttribute.h>
 #include <ImfFloatAttribute.h>
+#include <ImfFrameBuffer.h>
 #include <ImfHeader.h>
 #include <ImfInputFile.h>
 #include <ImfIntAttribute.h>
@@ -33,11 +62,28 @@ typedef int Py_ssize_t;
 #include <ImfVecAttribute.h>
 #include <ImfVersion.h>
 
+#include <ImfRationalAttribute.h>
+#include <ImfRational.h>
+#include <ImfKeyCodeAttribute.h>
+#include <ImfKeyCode.h>
+#include <ImfTimeCodeAttribute.h>
+#include <ImfTimeCode.h>
+
+#include <fstream>
 #include <iostream>
 #include <iomanip>
 #include <iostream>
 #include <vector>
 
+#define IMATH_VERSION IMATH_VERSION_MAJOR * 10000 + \
+                      IMATH_VERSION_MINOR * 100 + \
+                      IMATH_VERSION_PATCH
+
+#if IMATH_VERSION >= 30001
+#define Int64 uint64_t
+#define SInt64 int64_t
+#endif
+
 using namespace std;
 using namespace Imf;
 using namespace Imath;
@@ -52,6 +98,145 @@ static PyObject *PyObject_StealAttrString(PyObject* o, const char *name)
     return r;
 }
 
+static PyObject *PyObject_Call1(PyObject *f, PyObject* a)
+{
+    PyObject *r = PyObject_CallObject(f, a);
+    Py_DECREF(a);
+    return r;
+}
+
+////////////////////////////////////////////////////////////////////////
+//    Istream and Ostream derivatives
+////////////////////////////////////////////////////////////////////////
+
+class C_IStream: public IStream
+{
+  public:
+    C_IStream (PyObject *fo):
+        IStream(""), _fo(fo) {}
+    virtual bool    read (char c[], int n);
+    virtual Int64   tellg ();
+    virtual void    seekg (Int64 pos);
+    virtual void    clear ();
+    virtual const char*     fileName() const;
+  private:
+    PyObject *_fo;
+};
+
+bool
+C_IStream::read (char c[], int n)
+{
+    PyObject *data = PyObject_CallMethod(_fo, (char*)"read", (char*)"(i)", n);
+    if (data != NULL && PyString_AsString(data) && PyString_Size(data) == (Py_ssize_t)n) {
+      memcpy(c, PyString_AsString(data), PyString_Size(data));
+      Py_DECREF(data);
+    } else {
+      throw Iex::InputExc("file read failed");
+    }
+    return 0;
+}
+
+const char* C_IStream::fileName() const
+{
+  return "xxx";
+}
+
+
+Int64
+C_IStream::tellg ()
+{
+    PyObject *rv = PyObject_CallMethod(_fo, (char*)"tell", NULL);
+    if (rv != NULL && PyNumber_Check(rv)) {
+      PyObject *lrv = PyNumber_Long(rv);
+      long long t = PyLong_AsLong(lrv);
+      Py_DECREF(lrv);
+      Py_DECREF(rv);
+      return (Int64)t;
+    } else {
+      throw Iex::InputExc("tell failed");
+    }
+}
+
+void
+C_IStream::seekg (Int64 pos)
+{
+    PyObject *data = PyObject_CallMethod(_fo, (char*)"seek", (char*)"(L)", pos);
+    if (data != NULL) {
+        Py_DECREF(data);
+    } else {
+      throw Iex::InputExc("seek failed");
+    }
+}
+
+void
+C_IStream::clear ()
+{
+}
+
+////////////////////////////////////////////////////////////////////////
+
+class C_OStream: public OStream
+{
+  public:
+    C_OStream (PyObject *fo): OStream(""), _fo(fo) {}
+    virtual void    write (const char *c, int n);
+    virtual Int64   tellp ();
+    virtual void    seekp (Int64 pos);
+    virtual void    clear ();
+    virtual const char*     fileName() const;
+  private:
+    PyObject *_fo;
+};
+
+
+void
+C_OStream::write (const char*c, int n)
+{
+    PyObject *data = PyObject_CallMethod(_fo, (char*)"write", (char*)"(s#)", c, n);
+    if (data != NULL) {
+      Py_DECREF(data);
+    } else {
+      throw Iex::InputExc("file write failed");
+    }
+}
+
+const char* C_OStream::fileName() const
+{
+  return "xxx";
+}
+
+
+Int64
+C_OStream::tellp ()
+{
+    PyObject *rv = PyObject_CallMethod(_fo, (char*)"tell", NULL);
+    if (rv != NULL && PyNumber_Check(rv)) {
+      PyObject *lrv = PyNumber_Long(rv);
+      long long t = PyLong_AsLong(lrv);
+      Py_DECREF(lrv);
+      Py_DECREF(rv);
+      return (Int64)t;
+    } else {
+      throw Iex::InputExc("tell failed");
+    }
+}
+
+void
+C_OStream::seekp (Int64 pos)
+{
+    PyObject *data = PyObject_CallMethod(_fo, (char*)"seek", (char*)"(L)", pos);
+    if (data != NULL) {
+        Py_DECREF(data);
+    } else {
+      throw Iex::InputExc("seek failed");
+    }
+}
+
+void
+C_OStream::clear ()
+{
+}
+
 ////////////////////////////////////////////////////////////////////////
 //    InputFile
 ////////////////////////////////////////////////////////////////////////
@@ -59,6 +244,8 @@ static PyObject *PyObject_StealAttrString(PyObject* o, const char *name)
 typedef struct {
     PyObject_HEAD
     InputFile i;
+    PyObject *fo;
+    C_IStream *istream;
     int is_opened;
 } InputFileC;
 
@@ -99,13 +286,18 @@ static PyObject *channel(PyObject *self, PyObject *args, PyObject *kw)
 
     Imf::PixelType pt;
     if (pixel_type != NULL) {
+        if (PyObject_GetAttrString(pixel_type,"v") == NULL) {
+            return PyErr_Format(PyExc_TypeError, "Invalid PixelType object");
+        }
         pt = PixelType(PyLong_AsLong(PyObject_StealAttrString(pixel_type, "v")));
     } else {
         pt = channelPtr->type;
     }
 
-    int width  = dw.max.x - dw.min.x + 1;
-    int height = maxy - miny + 1;
+    int xSampling = channelPtr->xSampling;
+    int ySampling = channelPtr->ySampling;
+    int width  = (dw.max.x - dw.min.x + 1) / xSampling;
+    int height = (maxy - miny + 1) / ySampling;
 
     size_t typeSize;
     switch (pt) {
@@ -133,10 +325,10 @@ static PyObject *channel(PyObject *self, PyObject *args, PyObject *kw)
         size_t ystride = typeSize * width;
         frameBuffer.insert(cname,
                            Slice(pt,
-                                 pixels - dw.min.x * xstride - miny * ystride,
+                                 pixels - dw.min.x * xstride / xSampling - miny * ystride / ySampling,
                                  xstride,
                                  ystride,
-                                 1,1,
+                                 xSampling, ySampling,
                                  0.0));
         file->setFrameBuffer(frameBuffer);
         file->readPixels(miny, maxy);
@@ -194,7 +386,7 @@ static PyObject *channels(PyObject *self, PyObject *args, PyObject *kw)
     PyObject *item;
 
     while ((item = PyIter_Next(iterator)) != NULL) {
-      char *cname = PyString_AsString(item);
+      char *cname = PyUTF8_AsSstring(item);
       Channel *channelPtr = channels.findChannel(cname);
       if (channelPtr == NULL) {
           return PyErr_Format(PyExc_TypeError, "There is no channel '%s' in the image", cname);
@@ -283,6 +475,12 @@ static PyObject *dict_from_header(Header h)
     PyObject *pCFunc = PyObject_GetAttrString(pModuleImath, "Compression");
     PyObject *pCHFunc = PyObject_GetAttrString(pModuleImath, "chromaticity");
     PyObject *pCHSFunc = PyObject_GetAttrString(pModuleImath, "Chromaticities");
+    PyObject *pLevelMode = PyObject_GetAttrString(pModuleImath, "LevelMode");
+    PyObject *pLevelRoundingMode = PyObject_GetAttrString(pModuleImath, "LevelRoundingMode");
+    PyObject *pTileDescription = PyObject_GetAttrString(pModuleImath, "TileDescription");
+    PyObject *pRationalFunc = PyObject_GetAttrString(pModuleImath, "Rational");
+    PyObject *pKeyCodeFunc = PyObject_GetAttrString(pModuleImath, "KeyCode");
+    PyObject *pTimeCodeFunc = PyObject_GetAttrString(pModuleImath, "TimeCode");
 
     for (Header::ConstIterator i = h.begin(); i != h.end(); ++i) {
         const Attribute *a = &i.attribute();
@@ -303,10 +501,54 @@ static PyObject *dict_from_header(Header h)
             Py_DECREF(boxArgs);
             Py_DECREF(ptargs[0]);
             Py_DECREF(ptargs[1]);
+        } else if (const KeyCodeAttribute *ka = dynamic_cast <const KeyCodeAttribute *> (a)) {
+            PyObject *args = Py_BuildValue("iiiiiii",
+                                           ka->value().filmMfcCode(),
+                                           ka->value().filmType(),
+                                           ka->value().prefix(),
+                                           ka->value().count(),
+                                           ka->value().perfOffset(),
+                                           ka->value().perfsPerFrame(),
+                                           ka->value().perfsPerCount());
+                ob = PyObject_CallObject(pKeyCodeFunc, args);
+                Py_DECREF(args);
+        } else if (const TimeCodeAttribute *ta = dynamic_cast <const TimeCodeAttribute *> (a)) {
+                PyObject *args = Py_BuildValue("iiiiiiiiiiiiiiiiii",
+                                               ta->value().hours(),
+                                               ta->value().minutes(),
+                                               ta->value().seconds(),
+                                               ta->value().frame(),
+                                               ta->value().dropFrame(),
+                                               ta->value().colorFrame(),
+                                               ta->value().fieldPhase(),
+                                               ta->value().bgf0(),
+                                               ta->value().bgf1(),
+                                               ta->value().bgf2(),
+                                               ta->value().binaryGroup(1),
+                                               ta->value().binaryGroup(2),
+                                               ta->value().binaryGroup(3),
+                                               ta->value().binaryGroup(4),
+                                               ta->value().binaryGroup(5),
+                                               ta->value().binaryGroup(6),
+                                               ta->value().binaryGroup(7),
+                                               ta->value().binaryGroup(8));
+                ob = PyObject_CallObject(pTimeCodeFunc, args);
+                Py_DECREF(args);
+
+        } else if (const RationalAttribute *ra = dynamic_cast <const RationalAttribute *> (a)) {
+            PyObject *args = Py_BuildValue("ii", ra->value().n, ra->value().d);
+            ob = PyObject_CallObject(pRationalFunc, args);
+            Py_DECREF(args);
         } else if (const PreviewImageAttribute *pia = dynamic_cast <const PreviewImageAttribute *> (a)) {
             int size = pia->value().width() * pia->value().height() * 4;
-            PyObject *args = Py_BuildValue("iis#", pia->value().width(), pia->value().height(), (char*)pia->value().pixels(), size);
+#if PY_MAJOR_VERSION >= 3
+            const char fmt[] = "iiy#";
+#else
+            const char fmt[] = "iis#";
+#endif
+            PyObject *args = Py_BuildValue(fmt, pia->value().width(), pia->value().height(), (char*)pia->value().pixels(), size);
             ob = PyObject_CallObject(pPIFunc, args);
+
             Py_DECREF(args);
         } else if (const LineOrderAttribute *ta = dynamic_cast <const LineOrderAttribute *> (a)) {
             PyObject *args = PyTuple_Pack(1, PyInt_FromLong(ta->value()));
@@ -316,8 +558,7 @@ static PyObject *dict_from_header(Header h)
             PyObject *args = PyTuple_Pack(1, PyInt_FromLong(ta->value()));
             ob = PyObject_CallObject(pCFunc, args);
             Py_DECREF(args);
-        } else
-        if (const ChannelListAttribute *ta = dynamic_cast <const ChannelListAttribute *> (a)) {
+        } else if (const ChannelListAttribute *ta = dynamic_cast <const ChannelListAttribute *> (a)) {
             const ChannelList cl = ta->value();
             PyObject *CS = PyDict_New();
             for (ChannelList::ConstIterator j = cl.begin(); j != cl.end(); ++j) {
@@ -334,8 +575,7 @@ static PyObject *dict_from_header(Header h)
                 Py_DECREF(chanarg);
             }
             ob = CS;
-        } else
-        if (const FloatAttribute *ta = dynamic_cast <const FloatAttribute *> (a)) {
+        } else if (const FloatAttribute *ta = dynamic_cast <const FloatAttribute *> (a)) {
             ob = PyFloat_FromDouble(ta->value());
         } else if (const IntAttribute *ta = dynamic_cast <const IntAttribute *> (a)) {
             ob = PyInt_FromLong(ta->value());
@@ -343,11 +583,14 @@ static PyObject *dict_from_header(Header h)
             PyObject *args = Py_BuildValue("ff", ta->value().x, ta->value().y);
             ob = PyObject_CallObject(pV2FFunc, args);
             Py_DECREF(args);
-        } else
-        if (const StringAttribute *ta = dynamic_cast <const StringAttribute *> (a)) {
+        } else if (const StringAttribute *ta = dynamic_cast <const StringAttribute *> (a)) {
             ob = PyString_FromString(ta->value().c_str());
-        } else
-        if (const ChromaticitiesAttribute *ta = dynamic_cast<const ChromaticitiesAttribute *>(a)) {
+        } else if (const TileDescriptionAttribute *ta = dynamic_cast<const TileDescriptionAttribute *>(a)) {
+            const TileDescription td = ta->value();
+            PyObject *m = PyObject_Call1(pLevelMode, Py_BuildValue("(i)", td.mode));
+            PyObject *r = PyObject_Call1(pLevelRoundingMode, Py_BuildValue("(i)", td.roundingMode));
+            ob = PyObject_Call1(pTileDescription, Py_BuildValue("(iiNN)", td.xSize, td.ySize, m, r));
+        } else if (const ChromaticitiesAttribute *ta = dynamic_cast<const ChromaticitiesAttribute *>(a)) {
             const Chromaticities &ch(ta->value());
             PyObject *rgbwargs[4];
             rgbwargs[0] = Py_BuildValue("ff", ch.red[0], ch.red[1]);
@@ -391,6 +634,12 @@ static PyObject *dict_from_header(Header h)
     Py_DECREF(pPIFunc);
     Py_DECREF(pLOFunc);
     Py_DECREF(pCFunc);
+    Py_DECREF(pLevelMode);
+    Py_DECREF(pLevelRoundingMode);
+    Py_DECREF(pTileDescription);
+    Py_DECREF(pRationalFunc);
+    Py_DECREF(pKeyCodeFunc);
+    Py_DECREF(pTimeCodeFunc);
 
     return object;
 }
@@ -410,8 +659,8 @@ static PyObject *isComplete(PyObject *self, PyObject *args)
 /* Method table */
 static PyMethodDef InputFile_methods[] = {
   {"header", inheader, METH_VARARGS},
-  {"channel", (PyCFunction)channel, METH_KEYWORDS},
-  {"channels", (PyCFunction)channels, METH_KEYWORDS},
+  {"channel", (PyCFunction)channel, METH_VARARGS | METH_KEYWORDS},
+  {"channels", (PyCFunction)channels, METH_VARARGS | METH_KEYWORDS},
   {"close", inclose, METH_VARARGS},
   {"isComplete", isComplete, METH_VARARGS},
   {NULL, NULL},
@@ -420,36 +669,31 @@ static PyMethodDef InputFile_methods[] = {
 static void
 InputFile_dealloc(PyObject *self)
 {
+    InputFileC *object = ((InputFileC *)self);
+    if (object->fo)
+        Py_DECREF(object->fo);
     Py_DECREF(inclose(self, NULL));
     PyObject_Del(self);
 }
 
-static PyObject *
-InputFile_GetAttr(PyObject *self, char *attrname)
-{
-    return Py_FindMethod(InputFile_methods, self, attrname);
-}
-
 static PyObject *
 InputFile_Repr(PyObject *self)
 {
     //PyObject *result = NULL;
     char buf[50];
 
-
     sprintf(buf, "InputFile represented");
-    return PyString_FromString(buf);
+    return PyUnicode_FromString(buf);
 }
 
 static PyTypeObject InputFile_Type = {
-    PyObject_HEAD_INIT(&PyType_Type)
-    0,
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
     "OpenEXR.InputFile",
     sizeof(InputFileC),
     0,
     (destructor)InputFile_dealloc,
     0,
-    (getattrfunc)InputFile_GetAttr,
+    0,
     0,
     0,
     (reprfunc)InputFile_Repr,
@@ -467,10 +711,16 @@ static PyTypeObject InputFile_Type = {
     
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
 
+    "OpenEXR Input file object",
+
     0,
     0,
     0,
     0,
+    0,
+    0,
+
+    InputFile_methods
 
     /* the rest are NULLs */
 };
@@ -478,14 +728,33 @@ static PyTypeObject InputFile_Type = {
 int makeInputFile(PyObject *self, PyObject *args, PyObject *kwds)
 {
     InputFileC *object = ((InputFileC *)self);
-    char *filename;
+    PyObject *fo;
+    char *filename = NULL;
 
-    if (!PyArg_ParseTuple(args, "s:InputFile", &filename))
+    if (PyArg_ParseTuple(args, "O:InputFile", &fo)) {
+      if (PyString_Check(fo)) {
+          filename = PyString_AsString(fo);
+          object->fo = NULL;
+          object->istream = NULL;
+      } else if (PyUnicode_Check(fo)) {
+          filename = PyUTF8_AsSstring(fo);
+          object->fo = NULL;
+          object->istream = NULL;
+      } else {
+          object->fo = fo;
+          Py_INCREF(fo);
+          object->istream = new C_IStream(fo);
+      }
+    } else {
        return -1;
+    }
 
     try
     {
-        new(&object->i) InputFile(filename);
+        if (filename != NULL)
+          new(&object->i) InputFile(filename);
+        else
+          new(&object->i) InputFile(*object->istream);
     }
     catch (const std::exception &e)
     {
@@ -506,9 +775,17 @@ int makeInputFile(PyObject *self, PyObject *args, PyObject *kwds)
 typedef struct {
     PyObject_HEAD
     OutputFile o;
+    C_OStream *ostream;
+    PyObject *fo;
     int is_opened;
 } OutputFileC;
 
+static void releaseviews(std::vector<Py_buffer> &views)
+{
+    for (size_t i=0; i < views.size(); i++)
+        PyBuffer_Release(&views[i]);
+}
+
 static PyObject *outwrite(PyObject *self, PyObject *args)
 {
     OutputFile *file = &((OutputFileC *)self)->o;
@@ -517,18 +794,27 @@ static PyObject *outwrite(PyObject *self, PyObject *args)
     Box2i dw = file->header().dataWindow();
     int width = dw.max.x - dw.min.x + 1;
     int height = dw.max.y - dw.min.y + 1;
-	PyObject *pixeldata;
-	
+    PyObject *pixeldata;
+        
     if (!PyArg_ParseTuple(args, "O!|i:writePixels", &PyDict_Type, &pixeldata, &height))
        return NULL;
 
+    ssize_t currentScanLine = file->currentScanLine();
+    if (file->header().lineOrder() == DECREASING_Y) {
+        // With DECREASING_Y, currentScanLine() returns the maximum Y value of
+        // the window on the first call, and decrements at each scan line.
+        // We have to adjust to point to the correct address in the client buffer.
+        currentScanLine = dw.max.y - currentScanLine + dw.min.y;
+    }
+
     FrameBuffer frameBuffer;
+    std::vector<Py_buffer> views;
 
     const ChannelList &channels = file->header().channels();
     for (ChannelList::ConstIterator i = channels.begin();
          i != channels.end();
          ++i) {
-        PyObject *channel_spec = PyDict_GetItem(pixeldata, PyString_FromString(i.name()));
+        PyObject *channel_spec = PyDict_GetItem(pixeldata, PyUnicode_FromString(i.name()));
         if (channel_spec != NULL) {
             Imf::PixelType pt = i.channel().type;
             int typeSize = 4;
@@ -545,24 +831,44 @@ static PyObject *outwrite(PyObject *self, PyObject *args)
             default:
                 break;
             }
+            int xSampling = i.channel().xSampling;
+            int ySampling = i.channel().ySampling;
             int yStride = typeSize * width;
+            char *srcPixels;
+            ssize_t expectedSize = (height * yStride) / (xSampling * ySampling);
+            Py_ssize_t bufferSize;
 
-            if (!PyString_Check(channel_spec)) {
-                PyErr_Format(PyExc_TypeError, "Data for channel '%s' must be a string", i.name());
-                return NULL;
-            }
-            if (PyString_Size(channel_spec) != (height * yStride)) {
-                PyErr_Format(PyExc_TypeError, "Data for channel '%s' should have size %d but got %zu", i.name(), (height * yStride), PyString_Size(channel_spec));
+            if (PyString_Check(channel_spec)) {
+                bufferSize = PyString_Size(channel_spec);
+                srcPixels = PyString_AsString(channel_spec);
+            } else if (PyObject_CheckBuffer(channel_spec)) {
+                Py_buffer view;
+                if (PyObject_GetBuffer(channel_spec, &view, PyBUF_CONTIG_RO) != 0) {
+                    releaseviews(views);
+                    PyErr_Format(PyExc_TypeError, "Unsupported buffer structure for channel '%s'", i.name());
+                    return NULL;
+                }
+                views.push_back(view);
+                bufferSize = view.len;
+                srcPixels = (char*)view.buf;
+            } else {
+                releaseviews(views);
+                PyErr_Format(PyExc_TypeError, "Data for channel '%s' must be a string or support buffer protocol", i.name());
                 return NULL;
             }
 
-            char *srcPixels = PyString_AsString(channel_spec);
+            if (bufferSize != expectedSize) {
+                releaseviews(views);
+                PyErr_Format(PyExc_TypeError, "Data for channel '%s' should have size %zu but got %zu", i.name(), expectedSize, bufferSize);
+                return NULL;
+            }
 
             frameBuffer.insert(i.name(),                        // name
                 Slice(pt,                                       // type
-                      srcPixels - dw.min.x * typeSize - file->currentScanLine() * yStride,                         // base 
+                      srcPixels - dw.min.x * typeSize / xSampling - currentScanLine * yStride / ySampling,                         // base
                       typeSize,                                 // xStride
-                      yStride));                                // yStride
+                      yStride,                                  // yStride
+                      xSampling, ySampling));                   // subsampling
         }
     }
 
@@ -573,10 +879,11 @@ static PyObject *outwrite(PyObject *self, PyObject *args)
     }
     catch (const std::exception &e)
     {
+        releaseviews(views);
         PyErr_SetString(PyExc_IOError, e.what());
         return NULL;
     }
-
+    releaseviews(views);
     Py_RETURN_NONE;
 }
 
@@ -608,16 +915,13 @@ static PyMethodDef OutputFile_methods[] = {
 static void
 OutputFile_dealloc(PyObject *self)
 {
+    OutputFileC *object = ((OutputFileC *)self);
+    if (object->fo)
+        Py_DECREF(object->fo);
     Py_DECREF(outclose(self, NULL));
     PyObject_Del(self);
 }
 
-static PyObject *
-OutputFile_GetAttr(PyObject *self, char *attrname)
-{
-    return Py_FindMethod(OutputFile_methods, self, attrname);
-}
-
 static PyObject *
 OutputFile_Repr(PyObject *self)
 {
@@ -625,18 +929,17 @@ OutputFile_Repr(PyObject *self)
     char buf[50];
 
     sprintf(buf, "OutputFile represented");
-    return PyString_FromString(buf);
+    return PyUnicode_FromString(buf);
 }
 
 static PyTypeObject OutputFile_Type = {
-    PyObject_HEAD_INIT(&PyType_Type)
-    0,
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
     "OpenEXR.OutputFile",
     sizeof(OutputFileC),
     0,
     (destructor)OutputFile_dealloc,
     0,
-    (getattrfunc)OutputFile_GetAttr,
+    0,
     0,
     0,
     (reprfunc)OutputFile_Repr,
@@ -654,24 +957,47 @@ static PyTypeObject OutputFile_Type = {
     
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
 
+    "OpenEXR Output file object",
+
     0,
     0,
     0,
     0,
+    0,
+    0,
+
+    OutputFile_methods
 
     /* the rest are NULLs */
 };
 
 int makeOutputFile(PyObject *self, PyObject *args, PyObject *kwds)
 {
-    char *filename;
+    PyObject *fo;
     PyObject *header_dict;
 
-    if (!PyArg_ParseTuple(args, "sO!:OutputFile", &filename, &PyDict_Type, &header_dict))
-       return -1;
+    char *filename = NULL;
 
     OutputFileC *object = (OutputFileC *)self;
 
+    if (PyArg_ParseTuple(args, "OO!:OutputFile", &fo, &PyDict_Type, &header_dict)) {
+      if (PyString_Check(fo)) {
+          filename = PyString_AsString(fo);
+          object->fo = NULL;
+          object->ostream = NULL;
+      } else if (PyUnicode_Check(fo)) {
+          filename = PyUTF8_AsSstring(fo);
+          object->fo = NULL;
+          object->ostream = NULL;
+      } else {
+          object->fo = fo;
+          Py_INCREF(fo);
+          object->ostream = new C_OStream(fo);
+      }
+    } else {
+       return -1;
+    }
+
     Header header(64, 64);
 
     PyObject *pB2i = PyObject_GetAttrString(pModuleImath, "Box2i");
@@ -681,47 +1007,86 @@ int makeOutputFile(PyObject *self, PyObject *args, PyObject *kwds)
     PyObject *pCOMP = PyObject_GetAttrString(pModuleImath, "Compression");
     PyObject *pPI = PyObject_GetAttrString(pModuleImath, "PreviewImage");
     PyObject *pCH = PyObject_GetAttrString(pModuleImath, "Chromaticities");
+    PyObject *pTD = PyObject_GetAttrString(pModuleImath, "TileDescription");
+    PyObject *pRA = PyObject_GetAttrString(pModuleImath, "Rational");
+    PyObject *pKA = PyObject_GetAttrString(pModuleImath, "KeyCode");
+    PyObject *pTC = PyObject_GetAttrString(pModuleImath, "TimeCode");
 
     Py_ssize_t pos = 0;
     PyObject *key, *value;
 
     while (PyDict_Next(header_dict, &pos, &key, &value)) {
+        const char *ks = PyUTF8_AsSstring(key);
         if (PyFloat_Check(value)) {
-            header.insert(PyString_AsString(key), FloatAttribute(PyFloat_AsDouble(value)));
+            header.insert(ks, FloatAttribute(PyFloat_AsDouble(value)));
         }
         else if (PyInt_Check(value)) {
-            header.insert(PyString_AsString(key), IntAttribute(PyInt_AsLong(value)));
-        } else if (PyString_Check(value)) {
-            header.insert(PyString_AsString(key), StringAttribute(PyString_AsString(value)));
+            header.insert(ks, IntAttribute(PyInt_AsLong(value)));
+        } else if (PyBytes_Check(value)) {
+            header.insert(ks, StringAttribute(PyString_AsString(value)));
         } else if (PyObject_IsInstance(value, pB2i)) {
             Box2i box(V2i(PyLong_AsLong(PyObject_StealAttrString(PyObject_StealAttrString(value, "min"), "x")),
                           PyLong_AsLong(PyObject_StealAttrString(PyObject_StealAttrString(value, "min"), "y"))),
                       V2i(PyLong_AsLong(PyObject_StealAttrString(PyObject_StealAttrString(value, "max"), "x")),
                           PyLong_AsLong(PyObject_StealAttrString(PyObject_StealAttrString(value, "max"), "y"))));
-            header.insert(PyString_AsString(key), Box2iAttribute(box));
+            header.insert(ks, Box2iAttribute(box));
         } else if (PyObject_IsInstance(value, pB2f)) {
             Box2f box(V2f(PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "min"), "x")),
                           PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "min"), "y"))),
                       V2f(PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "max"), "x")),
                           PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "max"), "y"))));
-            header.insert(PyString_AsString(key), Box2fAttribute(box));
+            header.insert(ks, Box2fAttribute(box));
         } else if (PyObject_IsInstance(value, pPI)) {
             PreviewImage pi(PyLong_AsLong(PyObject_StealAttrString(value, "width")),
                             PyLong_AsLong(PyObject_StealAttrString(value, "height")),
                             (Imf::PreviewRgba *)PyString_AsString(PyObject_StealAttrString(value, "pixels")));
-            header.insert(PyString_AsString(key), PreviewImageAttribute(pi));
+            header.insert(ks, PreviewImageAttribute(pi));
         } else if (PyObject_IsInstance(value, pV2f)) {
             V2f v(PyFloat_AsDouble(PyObject_StealAttrString(value, "x")), PyFloat_AsDouble(PyObject_StealAttrString(value, "y")));
 
-            header.insert(PyString_AsString(key), V2fAttribute(v));
+            header.insert(ks, V2fAttribute(v));
         } else if (PyObject_IsInstance(value, pLO)) {
             LineOrder i = (LineOrder)PyInt_AsLong(PyObject_StealAttrString(value, "v"));
 
-            header.insert(PyString_AsString(key), LineOrderAttribute(i));
+            header.insert(ks, LineOrderAttribute(i));
+        } else if (PyObject_IsInstance(value, pTC)) {
+            TimeCode v(PyLong_AsLong(PyObject_StealAttrString(value,"hours")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"minutes")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"seconds")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"frame")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"dropFrame")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"colorFrame")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"fieldPhase")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"bgf0")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"bgf1")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"bgf2")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup1")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup2")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup3")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup4")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup5")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup6")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup7")),
+                       PyLong_AsLong(PyObject_StealAttrString(value,"binaryGroup8")));
+                                
+            header.insert(ks, TimeCodeAttribute(v));
+        
+        } else if (PyObject_IsInstance(value, pKA)) {
+                KeyCode v(PyLong_AsLong(PyObject_StealAttrString(value, "filmMfcCode")),
+                          PyLong_AsLong(PyObject_StealAttrString(value, "filmType")),
+                          PyLong_AsLong(PyObject_StealAttrString(value, "prefix")),
+                          PyLong_AsLong(PyObject_StealAttrString(value, "count")),
+                          PyLong_AsLong(PyObject_StealAttrString(value, "perfOffset")),
+                          PyLong_AsLong(PyObject_StealAttrString(value, "perfsPerFrame")),
+                          PyLong_AsLong(PyObject_StealAttrString(value, "perfsPerCount")));
+                header.insert(ks, KeyCodeAttribute(v));
+        } else if (PyObject_IsInstance(value, pRA)) {
+            Rational v(PyLong_AsLong(PyObject_StealAttrString(value, "n")), PyLong_AsLong(PyObject_StealAttrString(value, "d")));       
+            header.insert(ks, RationalAttribute(v));
         } else if (PyObject_IsInstance(value, pCOMP)) {
             Compression i = (Compression)PyInt_AsLong(PyObject_StealAttrString(value, "v"));
 
-            header.insert(PyString_AsString(key), CompressionAttribute(i));
+            header.insert(ks, CompressionAttribute(i));
         } else if (PyObject_IsInstance(value, pCH)) {
             V2f red(PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "red"), "x")),
                     PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "red"), "y")));
@@ -732,7 +1097,14 @@ int makeOutputFile(PyObject *self, PyObject *args, PyObject *kwds)
             V2f white(PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "white"), "x")),
                       PyFloat_AsDouble(PyObject_StealAttrString(PyObject_StealAttrString(value, "white"), "y")));
             Chromaticities c(red, green, blue, white);
-            header.insert(PyString_AsString(key), ChromaticitiesAttribute(c));
+            header.insert(ks, ChromaticitiesAttribute(c));
+        } else if (PyObject_IsInstance(value, pTD)) {
+            TileDescription td(PyInt_AsLong(PyObject_StealAttrString(value, "xSize")),
+                               PyInt_AsLong(PyObject_StealAttrString(value, "ySize")),
+                               (Imf::LevelMode)PyInt_AsLong(PyObject_StealAttrString(PyObject_StealAttrString(value, "mode"), "v")),
+                               (Imf::LevelRoundingMode)PyInt_AsLong(PyObject_StealAttrString(PyObject_StealAttrString(value, "roundingMode"), "v"))
+                               );
+            header.insert(ks, TileDescriptionAttribute(td));
         } else if (PyDict_Check(value)) {
             PyObject *key2, *value2;
             Py_ssize_t pos2 = 0;
@@ -742,7 +1114,7 @@ int makeOutputFile(PyObject *self, PyObject *args, PyObject *kwds)
                     printf("%s -> %s\n",
                         PyString_AsString(key2),
                         PyString_AsString(PyObject_Str(PyObject_Type(value2))));
-                header.channels().insert(PyString_AsString(key2),
+                header.channels().insert(PyUTF8_AsSstring(key2),
                                          Channel(PixelType(PyLong_AsLong(PyObject_StealAttrString(PyObject_StealAttrString(value2, "type"), "v"))),
                                                  PyLong_AsLong(PyObject_StealAttrString(value2, "xSampling")),
                                                  PyLong_AsLong(PyObject_StealAttrString(value2, "ySampling"))));
@@ -751,11 +1123,11 @@ int makeOutputFile(PyObject *self, PyObject *args, PyObject *kwds)
         } else if (PyList_Check(value)) {
             StringVector sv(PyList_Size(value));
             for (size_t i = 0; i < sv.size(); i++)
-                sv[i] = PyString_AsString(PyList_GetItem(value, i));
-            header.insert(PyString_AsString(key), StringVectorAttribute(sv));
+                sv[i] = PyUTF8_AsSstring(PyList_GetItem(value, i));
+            header.insert(ks, StringVectorAttribute(sv));
 #endif
         } else {
-            printf("XXX - unknown attribute: %s\n", PyString_AsString(PyObject_Str(key)));
+            printf("XXX - unknown attribute: %s\n", PyUTF8_AsSstring(PyObject_Str(key)));
         }
     }
 
@@ -765,10 +1137,18 @@ int makeOutputFile(PyObject *self, PyObject *args, PyObject *kwds)
     Py_DECREF(pLO);
     Py_DECREF(pCOMP);
     Py_DECREF(pPI);
+    Py_DECREF(pCH);
+    Py_DECREF(pTD);
+    Py_DECREF(pRA);
+    Py_DECREF(pKA);
+    Py_DECREF(pTC);
 
     try
     {
-        new(&object->o) OutputFile(filename, header);
+        if (filename != NULL)
+          new(&object->o) OutputFile(filename, header);
+        else
+          new(&object->o) OutputFile(*object->ostream, header);
     }
     catch (const std::exception &e)
     {
@@ -813,24 +1193,37 @@ PyObject *_isOpenExrFile(PyObject *self, PyObject *args)
     return PyBool_FromLong(isOpenExrFile(filename));
 }
 
+#ifdef VERSION_HAS_ISTILED
+PyObject *_isTiledOpenExrFile(PyObject *self, PyObject *args)
+{
+    char *filename;
+    if (!PyArg_ParseTuple(args, "s:isTiledOpenExrFile", &filename))
+        return NULL;
+    return PyBool_FromLong(isTiledOpenExrFile(filename));
+}
+#endif
+
 ////////////////////////////////////////////////////////////////////////
 
 static PyMethodDef methods[] = {
     {"Header", makeHeader, METH_VARARGS},
     {"isOpenExrFile", _isOpenExrFile, METH_VARARGS},
+#ifdef VERSION_HAS_ISTILED
+    {"isTiledOpenExrFile", _isTiledOpenExrFile, METH_VARARGS},
+#endif
     {NULL, NULL},
 };
 
-extern "C" void initOpenEXR()
+MOD_INIT(OpenEXR)
 {
     PyObject *m, *d, *item;
 
     Imf::staticInitialize();
 
-    m = Py_InitModule("OpenEXR", methods);
+    MOD_DEF(m, "OpenEXR", "", methods)
     d = PyModule_GetDict(m);
 
-    pModuleImath = PyImport_Import(item= PyString_FromString("Imath")); Py_DECREF(item);
+    pModuleImath = PyImport_ImportModule("Imath");
 
     /* initialize module variables/constants */
     InputFile_Type.tp_new = PyType_GenericNew;
@@ -838,9 +1231,9 @@ extern "C" void initOpenEXR()
     OutputFile_Type.tp_new = PyType_GenericNew;
     OutputFile_Type.tp_init = makeOutputFile;
     if (PyType_Ready(&InputFile_Type) != 0)
-        return;
+        return MOD_ERROR_VAL;
     if (PyType_Ready(&OutputFile_Type) != 0)
-        return;
+        return MOD_ERROR_VAL;
     PyModule_AddObject(m, "InputFile", (PyObject *)&InputFile_Type);
     PyModule_AddObject(m, "OutputFile", (PyObject *)&OutputFile_Type);
 
@@ -856,4 +1249,6 @@ extern "C" void initOpenEXR()
     PyDict_SetItemString(d, "HALF", item= PyLong_FromLong(HALF)); Py_DECREF(item);
     PyDict_SetItemString(d, "FLOAT", item= PyLong_FromLong(FLOAT)); Py_DECREF(item);
     PyDict_SetItemString(d, "__version__", item= PyString_FromString(VERSION)); Py_DECREF(item);
+
+    return MOD_SUCCESS_VAL(m);
 }
